@@ -6,16 +6,22 @@ import requests
 import logging
 import time
 import json
+import sys
 
 import esppy.espapi.tools as tools
 
-class K8S(object):
-    def __init__(self,url):
+from requests.packages.urllib3.exceptions import InsecureRequestWarning
+requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
+
+class K8S(tools.Options):
+    def __init__(self,url,esp = None,**kwargs):
+        tools.Options.__init__(self,**kwargs)
 
         self._url = urlparse(url)
+        self._esp = esp
         self._proxy = False
 
-        if self._url.scheme.index("-") != -1:
+        if self._url.scheme.find("-") != -1:
 
             a = self._url.scheme.split("-")
 
@@ -156,6 +162,23 @@ class K8S(object):
     def pod(self):
         return(self._pod)
 
+    def getNamespaces(self,delegate):
+        if tools.supports(delegate,"handleNamespaces") == False:
+            raise Exception("the delegate must implement the handleNamespaces function")
+
+        url = self.baseUrl
+        url += "api/v1/namespaces"
+
+        response = requests.get(url)
+
+        a = []
+
+        if response.status_code == 200:
+            data = response.json()
+            a = data["items"]
+
+        delegate.handleNamespaces(a)
+
     def getMyProjects(self,delegate):
         return(self.getProjects(delegate,namespace=self.namespace,name=self.project))
 
@@ -203,7 +226,7 @@ class K8S(object):
             if kind == "ESPServer":
                 a = [data]
             elif kind == "ESPServerList":
-                a = data.items
+                a = data["items"]
 
             delegate.handleProjects(a)
 
@@ -261,7 +284,7 @@ class K8S(object):
         pod = None
         pods = self.getPods()
         if len(pods) == 1:
-            pod = pods[0];
+            pod = pods[0]
         return(pod)
 
     def getLog(self,delegate):
@@ -307,96 +330,168 @@ class K8S(object):
         self.getPod(Tmp(self))
 
 class K8SProject(K8S):
-    def __init__(self,url):
-        K8S.__init__(self,url)
+    def __init__(self,url,esp,**kwargs):
+        K8S.__init__(self,url,esp,**kwargs)
 
         if self._project == None:
             raise Exception("URL must be in form protocol://server/<namespace>/<project>")
 
         self._config = None
 
+        model = None
+
+        if self.hasOpt("model_file"):
+            with open(self.getOpt("model_file")) as reader:
+                model = reader.read()
+        elif self.hasOpt("model_data"):
+            model = self.getOpt("model_data")
+
+        self.authenticate()
+
         if self.loadConfig() == False:
-            model = self.getDefaultModel()
+            if model == None:
+                model = self.getDefaultModel()
+
+        if model != None:
             self.load(model)
 
-    @property
-    def espUrl(self):
-        url = ""
-        if self._config != None:
-            if self.protocol == "k8s:" or self.protocol == "https:":
-                url += "https://"
-            elif self.protocol == "k8s-proxy:" or self.protocol == "https-proxy:":
-                url += "https://"
+    def authenticate(self):
+        while True:
+            status = self.getAuthToken()
+            if status == 401:
+                user = input("User: ")
+                pw = input("Password: ")
+                self.setOpt("username",user)
+                self.setOpt("password",pw)
             else:
-                url += "http://"
-            url += self._config["access"]["externalURL"]
-            url += "/SASEventStreamProcessingServer"
-            url += "/" + self._project
+                break
 
-        return(url)
+    def getAuthToken(self):
 
-    @property
-    def modelXml(self):
-        xml = ""
-        if self._config != None:
-            xml = self._config["spec"]["espProperties"]["server.xml"]
-        return(xml)
+        status = 0
 
-    def connect(self,connect,delegate,**kwargs):
-        opts = tools.Options(**kwargs)
-        modelconfig = opts.getOpt("model")
+        ingress = self.getIngress("sas-logon-app")
 
-        #k8s = self
+        if ingress is not None:
+            o = self.saslogon(ingress)
+            if o is not None:
+                status = o["status"]
+                if "token" in o:
+                    self.setOpt("viya",True);
+                    self.setOpt("access_token",o["token"])
+        else:
+            ingress = self.getIngress("oauth2-proxy")
+            if ingress is not None:
+                ingress = self.getIngress("uaa")
+                if ingress is not None:
+                    o = self.uaa(ingress)
+                    if o is not None:
+                        status = o["status"]
+                        if "token" in o:
+                            self.setOpt("access_token",o["token"])
 
-        class Loader(object):
-            def __init__(self):
-                pass
-        def loaded(self):
-            connect.connect(k8s.espUrl,delegate,options,start)
+        return(status)
 
-        class Handler(object):
-            def __init__(self,k8s):
-                self._k8s = k8s
+    def getIngress(self,name):
+        url = self.baseUrl
+        url += "apis/networking.k8s.io/v1beta1"
 
-            def loaded(self):
-                if modelconfig == None:
-                    class Tmp(object):
-                        def __init__(self,k8s):
-                            self._k8s = k8s
-                        def handlePod(self,pod):
-                            self._k8s._pod = pod
-                            connect.connect(self._k8s.espUrl,delegate,**kwargs)
-                    self._k8s.getPod(Tmp(self._k8s))
-                else:
-                    class Tmp(object):
-                        def __init__(self):
-                            pass
-                        def modelHandler(self,model):
-                            k8s.load(model,opts.getOpts(),Loader())
-                    k8s.getModel(modelconfig,Tmp())
+        if self.namespace is not None:
+            url += "/namespaces/" + self.namespace
+        url += "/ingresses/" + name
+        response = requests.get(url)
+        if response.status_code >= 403:
+            return(None)
+        else:
+            return(json.loads(response.text))
 
-            def notFound(self,project):
-                if modelconfig == None:
-                    if opts.getOpt("create",True):
-                        model = k8s.getDefaultModel(project)
-                        k8s.load(model,opts.getOpts(),Loader())
-                    elif tools.supports(delegate,"error"):
-                        delegate.error("project not found: " + project)
-                else:
-                    class Tmp(object):
-                        def __init__(self):
-                            pass
-                        def modelHandler(self,model):
-                            k8s.load(model,opts.getOpts(),Loader())
+    def saslogon(self,data):
 
-                    k8s.getModel(modelconfig,Tmp())
+        secret = self.getSecret()
 
-            def error(self,request,error):
-                raise Eexception("error: " + opts)
+        url = "https://";
+        url += data["spec"]["tls"][0]["hosts"][0];
+        url += "/SASLogon/oauth/clients/consul";
+        url += "?callback=false&serviceId=app";
 
-        self.loadConfig(Handler(self))
+        ca_bundle = self._esp.ca_bundle
 
-    def loadConfig(self,delegate = None):
+        if ca_bundle != None:
+            if ca_bundle == "_noverify_":
+                response = requests.post(url,headers={"X-Consul-Token":secret},verify=False);
+            else:
+                response = requests.post(url,headers={"X-Consul-Token":secret},verify=ca_bundle);
+        else:
+            response = requests.post(url,headers={"X-Consul-Token":secret});
+
+        if response.status_code >= 400:
+            return(None)
+
+        o = json.loads(response.text)
+
+        saslogon = {"status":response.status_code,"token":o["access_token"]}
+
+        return(saslogon)
+
+    def uaa(self,data):
+        url = "https://"
+        url += data["spec"]["tls"][0]["hosts"][0]
+        url += "/uaa/oauth/token"
+
+        user = self.getOpt("username","")
+        pw = self.getOpt("password","")
+
+        send = ""
+        send += "client_id=sv_client"
+        send += "&client_secret=secret"
+        send += "&grant_type=password"
+        send += "&username=" + user
+        send += "&password=" + pw
+
+        ca_bundle = self._esp.ca_bundle
+
+        if ca_bundle != None:
+            if ca_bundle == "_noverify_":
+                response = requests.post(url,data=send,headers={"Content-Type":"application/x-www-form-urlencoded","Accept":"application/json"},verify=False)
+            else:
+                response = requests.post(url,data=send,headers={"Content-Type":"application/x-www-form-urlencoded","Accept":"application/json"},verify=ca_bundle)
+        else:
+            response = requests.post(url,data=send,headers={"Content-Type":"application/x-www-form-urlencoded","Accept":"application/json"})
+
+        if response.status_code >= 400:
+            return({"status":response.status_code})
+
+        o = json.loads(response.text)
+
+        uaa = {"status":response.status_code,"token":o["access_token"]}
+
+        return(uaa)
+
+    def getSecret(self):
+
+        secret = None
+
+        url = self.baseUrl;
+        url += "api/v1";
+
+        if self.namespace is not None:
+            url += "/namespaces/" + self.namespace
+
+        url += "/secrets/sas-consul-client";
+
+        response = requests.get(url)
+        if response.status_code >= 404:
+            return(None)
+        else:
+            o = json.loads(response.text)
+            if "code" in o and o["code"] == 404:
+                return(None)
+            else:
+                secret = b64decode(o["data"]["CONSUL_HTTP_TOKEN"]).decode()
+
+        return(secret)
+
+    def loadConfig(self):
 
         self._config = None
 
@@ -408,7 +503,15 @@ class K8SProject(K8S):
         url += "/espservers"
         url += "/" + self._project
 
-        response = requests.get(url)
+        ca_bundle = self._esp.ca_bundle
+
+        if ca_bundle != None:
+            if ca_bundle == "_noverify_":
+                response = requests.get(url,verify=False)
+            else:
+                response = requests.get(url,verify=ca_bundle)
+        else:
+            response = requests.get(url)
 
         code = False
 
@@ -432,6 +535,7 @@ class K8SProject(K8S):
         newmodel = "b64" + b64encode(model.encode("utf-8")).decode()
 
         if newmodel == self.modelXml:
+        #if False:
             if opts.getOpt("overwrite",False) == False and opts.getOpt("force",False) == False:
                 return
 
@@ -446,7 +550,7 @@ class K8SProject(K8S):
         url += "/espservers/"
         url += self._project
 
-        content = self.getYaml(newmodel)
+        content = self.getYaml(newmodel,pv = False)
         headers = {"content-type":"application/yaml","accept":"application/json"}
 
         response = requests.post(url,data=content,headers=headers)
@@ -467,7 +571,10 @@ class K8SProject(K8S):
 
         response = requests.delete(url)
 
-    def getYaml(self,model):
+    def getYaml(self,model,**kwargs):
+
+        opts = tools.Options(**kwargs)
+
         s = ""
 
         s += "apiVersion: iot.sas.com/v1alpha1\n"
@@ -499,10 +606,13 @@ class K8SProject(K8S):
         s += "            matchLabels:\n"
         s += "          template:\n"
         s += "            spec:\n"
-        s += "               volumes:\n"
-        s += "               - name: data\n"
-        s += "                 persistentVolumeClaim:\n"
-        s += "                   claimName: esp-pv\n"
+
+        if (self.getOpt("viya",False) == False):
+            s += "               volumes:\n"
+            s += "               - name: data\n"
+            s += "                 persistentVolumeClaim:\n"
+            s += "                   claimName: esp-pv\n"
+
         s += "               containers:\n"
         s += "               - name: ((PROJECT_SERVICE_NAME))\n"
         s += "                 resources:\n"
@@ -512,9 +622,12 @@ class K8SProject(K8S):
         s += "                   limits:\n"
         s += "                     memory: \"2Gi\"\n"
         s += "                     cpu: \"2\"\n"
-        s += "                 volumeMounts:\n"
-        s += "                 - mountPath: /mnt/data\n"
-        s += "                   name: data\n"
+
+        if (self.getOpt("viya",False) == False):
+            s += "                 volumeMounts:\n"
+            s += "                 - mountPath: /mnt/data\n"
+            s += "                   name: data\n"
+
         s += "    loadBalancerTemplate:\n"
         s += "      deployment:\n"
         s += "        spec:\n"
@@ -561,12 +674,24 @@ class K8SProject(K8S):
 
             if state == "Succeeded":
                 pod = self.getPod()
-                status = pod["status"]["phase"]
-                if status == "Running":
+                status = pod["status"]
+                phase = status["phase"]
+
+                if "containerStatuses" in status:
+                    containerStatus = status["containerStatuses"][0];
+
+                    if "lastState" in containerStatus:
+                        lastState = containerStatus["lastState"]
+
+                        if "terminated" in lastState:
+                            if lastState["terminated"]["exitCode"] != 0 or lastState["terminated"]["reason"] == "Error":
+                                raise Exception(json.dumps(containerStatus,indent=2))
+
+                if phase == "Running":
                     conditions = pod["status"]["conditions"]
                     ready = True
                     for i in range(0,len(conditions)):
-                        condition = conditions[i];
+                        condition = conditions[i]
                         if condition["status"] != "True":
                             ready = False
                             break
@@ -578,8 +703,8 @@ class K8SProject(K8S):
         self.readiness()
 
     def readiness(self):
-        url = self.espUrl;
-        url += "/internal/ready";
+        url = self.espUrl
+        url += "/internal/ready"
 
         success = False
 
@@ -593,14 +718,37 @@ class K8SProject(K8S):
 
             time.sleep(1)
 
-def create(url):
+    @property
+    def espUrl(self):
+        url = ""
+        if self._config != None:
+            if self.protocol == "k8s:" or self.protocol == "https:":
+                url += "https://"
+            elif self.protocol == "k8s-proxy:" or self.protocol == "https-proxy:":
+                url += "https://"
+            else:
+                url += "http://"
+            url += self._config["access"]["externalURL"]
+            url += "/SASEventStreamProcessingServer"
+            url += "/" + self._project
+
+        return(url)
+
+    @property
+    def modelXml(self):
+        xml = ""
+        if self._config != None:
+            xml = self._config["spec"]["espProperties"]["server.xml"]
+        return(xml)
+
+def create(url,esp = None,**kwargs):
     u = urlparse(url)
 
     o = None
 
     if u.path != None and len(u.path.split("/")) == 3:
-        o = K8SProject(url)
+        o = K8SProject(url,esp,**kwargs)
     else:
-        o = K8S(url)
+        o = K8S(url,esp,**kwargs)
 
     return(o)
